// AniCursor.java

// Create an animated cursor based on a Microsoft .ani file.

package ca.mb.javajeff.anicursor;

import java.awt.*;
import java.awt.image.*;

import java.io.*;

import java.util.*;

import javax.swing.*;

import ca.mb.javajeff.cur.*;
import ca.mb.javajeff.ico.*;
import ca.mb.javajeff.riff.*;

public class AniCursor implements Runnable
{
   private volatile boolean finished; // Animation completion flag

   private volatile Component comp;

   private volatile Cursor [] cursors;

   private int [] durations;

   private int [] sequence;

   private int index;

   public AniCursor (String aniName, Component comp)
      throws BadAniException, IOException
   {
      this.comp = comp;

      if (aniName == null)
      {
          cursors = new Cursor [8];
          cursors [0] = Cursor.getPredefinedCursor (Cursor.N_RESIZE_CURSOR);
          cursors [1] = Cursor.getPredefinedCursor (Cursor.NE_RESIZE_CURSOR);
          cursors [2] = Cursor.getPredefinedCursor (Cursor.E_RESIZE_CURSOR);
          cursors [3] = Cursor.getPredefinedCursor (Cursor.SE_RESIZE_CURSOR);
          cursors [4] = Cursor.getPredefinedCursor (Cursor.S_RESIZE_CURSOR);
          cursors [5] = Cursor.getPredefinedCursor (Cursor.SW_RESIZE_CURSOR);
          cursors [6] = Cursor.getPredefinedCursor (Cursor.W_RESIZE_CURSOR);
          cursors [7] = Cursor.getPredefinedCursor (Cursor.NW_RESIZE_CURSOR);

          durations = new int [8];
          for (int i = 0; i < durations.length; i++)
               durations [i] = 75;

          sequence = new int [8];
          for (int i = 0; i < sequence.length; i++)
               sequence [i] = i;
      }
      else
      {
          RIFF riff = null;
          try
          {
              riff = new RIFF (aniName);
              if (!riff.getFormType ().equals ("ACON"))
                  throw new BadAniException ("not an animated cursor file");

              ArrayList<BufferedImage> bilist;
              bilist = new ArrayList<BufferedImage> ();

              int x;
              int y;

              Chunk chunk;
              while ((chunk = riff.getChunk ()) != null)
                 if (chunk.name.equals ("anih"))
                 {
                     int size = ubyte (chunk.data [3]);
                     size <<= 8;
                     size |= ubyte (chunk.data [2]);
                     size <<= 8;
                     size |= ubyte (chunk.data [1]);
                     size <<= 8;
                     size |= ubyte (chunk.data [0]);

                     int nframes = ubyte (chunk.data [7]);
                     nframes <<= 8;
                     nframes |= ubyte (chunk.data [6]);
                     nframes <<= 8;
                     nframes |= ubyte (chunk.data [5]);
                     nframes <<= 8;
                     nframes |= ubyte (chunk.data [4]);

                     int nsteps = ubyte (chunk.data [11]);
                     nsteps <<= 8;
                     nsteps |= ubyte (chunk.data [10]);
                     nsteps <<= 8;
                     nsteps |= ubyte (chunk.data [9]);
                     nsteps <<= 8;
                     nsteps |= ubyte (chunk.data [8]);

                     int width = ubyte (chunk.data [15]);
                     width <<= 8;
                     width |= ubyte (chunk.data [14]);
                     width <<= 8;
                     width |= ubyte (chunk.data [13]);
                     width <<= 8;
                     width |= ubyte (chunk.data [12]);

                     int height = ubyte (chunk.data [19]);
                     height <<= 8;
                     height |= ubyte (chunk.data [18]);
                     height <<= 8;
                     height |= ubyte (chunk.data [17]);
                     height <<= 8;
                     height |= ubyte (chunk.data [16]);

                     int bitcount = ubyte (chunk.data [23]);
                     bitcount <<= 8;
                     bitcount |= ubyte (chunk.data [22]);
                     bitcount <<= 8;
                     bitcount |= ubyte (chunk.data [21]);
                     bitcount <<= 8;
                     bitcount |= ubyte (chunk.data [20]);

                     int nplanes = ubyte (chunk.data [27]);
                     nplanes <<= 8;
                     nplanes |= ubyte (chunk.data [26]);
                     nplanes <<= 8;
                     nplanes |= ubyte (chunk.data [25]);
                     nplanes <<= 8;
                     nplanes |= ubyte (chunk.data [24]);

                     int disprate = ubyte (chunk.data [31]);
                     disprate <<= 8;
                     disprate |= ubyte (chunk.data [30]);
                     disprate <<= 8;
                     disprate |= ubyte (chunk.data [29]);
                     disprate <<= 8;
                     disprate |= ubyte (chunk.data [28]);

                     int attr = ubyte (chunk.data [35]);
                     attr <<= 8;
                     attr |= ubyte (chunk.data [34]);
                     attr <<= 8;
                     attr |= ubyte (chunk.data [33]);
                     attr <<= 8;
                     attr |= ubyte (chunk.data [32]);

                     if ((attr & 1) == 0)
                         throw new BadAniException ("cannot deal with raw "+
                                                    "bitmap data");
                     durations = new int [nsteps];
                     for (int i = 0; i < durations.length; i++)
                          durations [i] = disprate*1000/60; // convert from
                                                            // jiffies to
                                                            // milliseconds

                     sequence = new int [nsteps];
                     for (int i = 0; i < sequence.length; i++)
                          sequence [i] = i;
                 }
                 else
                 if (chunk.name.equals ("icon"))
                 {
                     BufferedImage bi;
                     int ncolors;

                     if (chunk.data [2] == 1 && chunk.data [3] == 0)
                     {
                         Ico i;
                         try
                         {
                             i = new Ico (new ByteArrayInputStream (chunk.data));
                         }
                         catch (BadIcoResException bire)
                         {
                             throw new BadAniException (bire);
                         }
                         bi = i.getImage (0);
                         ncolors = i.getNumColors (0);
                         x = 0;
                         y = 0;
                     }
                     else
                     if (chunk.data [2] == 2 && chunk.data [3] == 0)
                     {
                         Cur c;
                         try
                         {
                             c = new Cur (new ByteArrayInputStream (chunk.data));
                         }
                         catch (BadCurResException bcre)
                         {
                             throw new BadAniException (bcre);
                         }
                         bi = c.getImage (0);
                         ncolors = c.getNumColors (0);
                         x = c.getHotspotX (0);
                         y = c.getHotspotY (0);
                     }
                     else
                        throw new BadAniException ("icon or cursor resource "+
                                                   "expected");

                     // Convert a translucent alpha channel, where alpha
                     // ranges from 0 to 255 to a binary alpha channel, where
                     // alpha is either 0 or 255.

                     if (ncolors == 0)
                         for (int i = 0; i < bi.getHeight (); i++)
                         {
                              int [] rgb = bi.getRGB (0, i, bi.getWidth (), 1,
                                                      null, 0,
                                                      bi.getWidth ()*4);
                              for (int j = 0; j < rgb.length; j++)
                              {
                                   int alpha = (rgb [j] >> 24) & 255;
                                   if (alpha < 0x80)
                                       alpha = 0;
                                   else
                                       alpha = 255;
                                   rgb [j] &= 0x00ffffff;
                                   rgb [j] = (alpha << 24) | rgb [j];
                              }
                              bi.setRGB (0, i, bi.getWidth (), 1, rgb, 0,
                                         bi.getWidth ()*4);
                         }
                     bilist.add (bi);

                     cursors = new Cursor [bilist.size ()];

                     Toolkit toolkit = Toolkit.getDefaultToolkit ();
                     for (int i = 0; i < bilist.size (); i++)
                          cursors [i] = toolkit.
                                        createCustomCursor (bilist.get (i),
                                                            new Point (x, y),
                                                            "anicursor");
                 }
                 else
                 if (chunk.name.equals ("rate"))
                 {
                     if (durations == null)
                         throw new BadAniException ("ANI header missing");

                     for (int i = 0; i < durations.length; i++)
                     {
                          int rate = ubyte (chunk.data [4*i+3]);
                          rate <<= 8;
                          rate |= ubyte (chunk.data [4*i+2]);
                          rate <<= 8;
                          rate |= ubyte (chunk.data [4*i+1]);
                          rate <<= 8;
                          rate |= ubyte (chunk.data [4*i]);

                          durations [i] = rate*1000/60; // convert from jiffies
                                                        // to milliseconds
                     }
                 }
                 else
                 if (chunk.name.equals ("seq "))
                 {
                     if (sequence == null)
                         throw new BadAniException ("ANI header missing");

                     for (int i = 0; i < sequence.length; i++)
                     {
                          int index = ubyte (chunk.data [4*i+3]);
                          index <<= 8;
                          index |= ubyte (chunk.data [4*i+2]);
                          index <<= 8;
                          index |= ubyte (chunk.data [4*i+1]);
                          index <<= 8;
                          index |= ubyte (chunk.data [4*i]);

                          sequence [i] = index;
                     }
                 }
          }
          catch (BadRIFFException briffe)
          {
             throw new BadAniException (briffe);
          }
          finally
          {
             if (riff != null)
                 riff.close ();
          }

          if (durations == null)
              throw new BadAniException ("ANI header missing");
      }
   }

   public void run ()
   {
      index = 0;

      Runnable r = new Runnable ()
                   {
                        public void run ()
                        {
                           comp.setCursor (cursors [sequence [index]]);
                        }
                   };

      while (!finished)
      {
         try
         {
             Thread.currentThread ().sleep (durations [index]);
             SwingUtilities.invokeAndWait (r);
             index++;
             if (index == durations.length)
                 index = 0;
         }
         catch (Exception ex)
         {
         }
      }

      try
      {
          r = new Runnable ()
              {
                  public void run ()
                  {
                     comp.setCursor (Cursor.
                                     getPredefinedCursor (Cursor.
                                                          DEFAULT_CURSOR));
                  }
              };
          SwingUtilities.invokeAndWait (r);
      }
      catch (Exception ex)
      {
      }
   }

   public void start ()
   {
      finished = false;
      new Thread (this).start ();
   }

   public void stop ()
   {
      finished = true;
   }

   private int ubyte (byte b)
   {
      return (b < 0) ? 256+b : b; // Convert byte to unsigned byte.
   }
}
